/*
 * C code
 * File: dsdp_4vars.c
 *
 * This program calculates p-values for the sharpened Diks-Panchenko statistic
 * (sharpened assuming the Hall and Minotte (JRSS B, 2002) method) for
 * 4 variables, X, Y, Q and U
 *
 * Diks, C. and Wolski, M. (2016).
 * Nonlinear Granger causality: Guidelines for multivariate analysis.
 *  Journal of Applied Econometrics, 31, 1333-1351.
 *  DOI: 10.1002/jae.2495.

 * Code by: Cees Diks and Marcin Wolski -- October 2013
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#define max(a,b)	a>b?a:b

#define EPS             0.81 /* optimal bandwidth for n=632*/

int Ndat=15000;
int Mmax=1;     /* maximum emb. dim. */
int n, K;
double T2,Q,Qad,**A,C[4],*h,*ohm,*cov;
double p_HJ, p_T2, p_HJad, HJ_TVAL, T2_TVAL;

/* sharpening function - bias reduction to order eps^4 */
double phihat(double *z, double *x, double *y, double *q, double *u, char *what, char *where, int point, int N, double epsilon, double c)
{
	int i;
	double nominator, denominator, new;

	nominator=denominator=new=0.0;

	if(strcmp(where,"yqu")==0) {
		if(strcmp(what,"y")==0) {
			for (i=0;i!=N;i++)
			 nominator += (y[i]-y[point])*exp((-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			for (i=0;i!=N;i++)
			 denominator += exp((-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			new=y[point]+c*nominator/denominator;
		}
		if(strcmp(what,"q")==0) {
			for (i=0;i!=N;i++)
			 nominator += (q[i]-q[point])*exp((-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			for (i=0;i!=N;i++)
			 denominator += exp((-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			new=q[point]+c*nominator/denominator;
		}
		if(strcmp(what,"u")==0) {
			for (i=0;i!=N;i++)
			 nominator += (u[i]-u[point])*exp((-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			for (i=0;i!=N;i++)
			 denominator += exp((-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			new=u[point]+c*nominator/denominator;
		}
	}
	if(strcmp(where,"xyqu")==0) {
		if(strcmp(what,"x")==0) {
			for (i=0;i!=N;i++)
			 nominator += (x[i]-x[point])*exp((-(x[i]-x[point])*(x[i]-x[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			for (i=0;i!=N;i++)
			 denominator += exp((-(x[i]-x[point])*(x[i]-x[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			new=x[point]+c*nominator/denominator;
		}
		if(strcmp(what,"y")==0) {
			for (i=0;i!=N;i++)
			 nominator += (y[i]-y[point])*exp((-(x[i]-x[point])*(x[i]-x[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			for (i=0;i!=N;i++)
			 denominator += exp((-(x[i]-x[point])*(x[i]-x[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			new=y[point]+c*nominator/denominator;
		}
		if(strcmp(what,"q")==0) {
			for (i=0;i!=N;i++)
			 nominator += (q[i]-q[point])*exp((-(x[i]-x[point])*(x[i]-x[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			for (i=0;i!=N;i++)
			 denominator += exp((-(x[i]-x[point])*(x[i]-x[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			new=q[point]+c*nominator/denominator;
		}
		if(strcmp(what,"u")==0) {
			for (i=0;i!=N;i++)
			 nominator += (u[i]-u[point])*exp((-(x[i]-x[point])*(x[i]-x[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			for (i=0;i!=N;i++)
			 denominator += exp((-(x[i]-x[point])*(x[i]-x[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			new=u[point]+c*nominator/denominator;
		}
	}
	if(strcmp(where,"yzqu")==0) {
		if(strcmp(what,"z")==0) {
			for (i=0;i!=N;i++)
			 nominator += (z[i]-z[point])*exp((-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			for (i=0;i!=N;i++)
			 denominator += exp((-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			new=z[point]+c*nominator/denominator;
		}
		if(strcmp(what,"y")==0) {
			for (i=0;i!=N;i++)
			 nominator += (y[i]-y[point])*exp((-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));
			for (i=0;i!=N;i++)
			 denominator += exp((-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			new=y[point]+c*nominator/denominator;
		}
		if(strcmp(what,"q")==0) {
			for (i=0;i!=N;i++)
			 nominator += (q[i]-q[point])*exp((-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			for (i=0;i!=N;i++)
			 denominator += exp((-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			new=q[point]+c*nominator/denominator;
		}
		if(strcmp(what,"u")==0) {
			for (i=0;i!=N;i++)
			 nominator += (u[i]-u[point])*exp((-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			for (i=0;i!=N;i++)
			 denominator += exp((-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			new=u[point]+c*nominator/denominator;
		}

	}
	if(strcmp(where,"xyzqu")==0) {
		if(strcmp(what,"x")==0) {
			for (i=0;i!=N;i++)
			 nominator += (x[i]-x[point])*exp((-(x[i]-x[point])*(x[i]-x[point])-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			for (i=0;i!=N;i++)
			 denominator += exp((-(x[i]-x[point])*(x[i]-x[point])-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			new=x[point]+c*nominator/denominator;
		}
		if(strcmp(what,"y")==0) {
			for (i=0;i!=N;i++)
			 nominator += (y[i]-y[point])*exp((-(x[i]-x[point])*(x[i]-x[point])-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			for (i=0;i!=N;i++)
			 denominator += exp((-(x[i]-x[point])*(x[i]-x[point])-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			new=y[point]+c*nominator/denominator;
		}
		if(strcmp(what,"z")==0) {
			for (i=0;i!=N;i++)
			 nominator += (z[i]-z[point])*exp((-(x[i]-x[point])*(x[i]-x[point])-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			for (i=0;i!=N;i++)
			 denominator += exp((-(x[i]-x[point])*(x[i]-x[point])-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			new=z[point]+c*nominator/denominator;
		}
		if(strcmp(what,"q")==0) {
			for (i=0;i!=N;i++)
			 nominator += (q[i]-q[point])*exp((-(x[i]-x[point])*(x[i]-x[point])-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			for (i=0;i!=N;i++)
			 denominator += exp((-(x[i]-x[point])*(x[i]-x[point])-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			new=q[point]+c*nominator/denominator;
		}
		if(strcmp(what,"u")==0) {
			for (i=0;i!=N;i++)
			 nominator += (u[i]-u[point])*exp((-(x[i]-x[point])*(x[i]-x[point])-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			for (i=0;i!=N;i++)
			 denominator += exp((-(x[i]-x[point])*(x[i]-x[point])-(z[i]-z[point])*(z[i]-z[point])-(y[i]-y[point])*(y[i]-y[point])-(q[i]-q[point])*(q[i]-q[point])-(u[i]-u[point])*(u[i]-u[point]))/(epsilon*epsilon));

			new=u[point]+c*nominator/denominator;
		}
	}
	return(new);
}

void redunGaussian(double *z, double *x, double *y, double *q, double *u, int N, int m, int mmax, double epsilon)
{

  int i, j, s, Const, powerK;
  int IYQij, IXYQij, IYZQij, IXYZQij;
  double eps2, C, disx, disy, disq, disz, *Cyq, *Cxyq, *Cyzq, *Cxyzq, disx_xyq, disx_xyzq, disz_yzq, disz_xyzq, disy_yq, disy_xyq, disy_yzq, disy_xyzq, disq_yq, disq_xyq, disq_yzq, disq_xyzq;
  /* the q vector is in fact qu vector in the 4 variable setting */

  /* determine the scaling constant */
  Const = 3;
  powerK=2;
  C=1;
  eps2=(1/sqrt(2*C))*epsilon;

  Cyq = (double *) malloc(N*sizeof(double));
  Cxyq = (double *) malloc(N*sizeof(double));
  Cyzq = (double *) malloc(N*sizeof(double));
  Cxyzq = (double *) malloc(N*sizeof(double));

  for (i=0;i!=N;i++)
    h[i] = Cyq[i] = Cxyq[i] = Cyzq[i] = Cxyzq[i] = 0.0;

  T2=0.0;

  n = N - mmax;

  for (i=mmax;i!=N;i++)
  {

    Cyq[i]=Cxyq[i]=Cyzq[i]=Cxyzq[i]=0.0;

    for (j=mmax;j!=N;j++)
    if (j!=i)
    {

      disx = disy = disq = disz = 0.0;
      disx_xyq = disx_xyzq = disz_yzq = disz_xyzq = disy_yq = disy_xyq = disy_yzq = disy_xyzq = disq_yq = disq_xyq = disq_yzq = disq_xyzq = 0.0;
      for (s=1;s!=m+1;s++) {
		/* vector YQ */
		disy_yq += pow((y[i-s]-phihat(z,x,y,q,u,"y","yqu",j-s,N,eps2,C)),powerK);
		disq_yq += pow((q[i-s]-phihat(z,x,y,q,u,"q","yqu",j-s,N,eps2,C)),powerK)+pow((u[i-s]-phihat(z,x,y,q,u,"u","yqu",j-s,N,eps2,C)),powerK);
		/* vector XYQ */
		disx_xyq += pow((x[i-s]-phihat(z,x,y,q,u,"x","xyqu",j-s,N,eps2,C)),powerK);
		disy_xyq += pow((y[i-s]-phihat(z,x,y,q,u,"y","xyqu",j-s,N,eps2,C)),powerK);
		disq_xyq += pow((q[i-s]-phihat(z,x,y,q,u,"q","xyqu",j-s,N,eps2,C)),powerK)+pow((u[i-s]-phihat(z,x,y,q,u,"u","xyqu",j-s,N,eps2,C)),powerK);
		/* vector YZQ */
		disy_yzq += pow((y[i-s]-phihat(z,x,y,q,u,"y","yzqu",j-s,N,eps2,C)),powerK);
		disz_yzq += pow((z[i-s]-phihat(z,x,y,q,u,"z","yzqu",j-s,N,eps2,C)),powerK);
		disq_yzq += pow((q[i-s]-phihat(z,x,y,q,u,"q","yzqu",j-s,N,eps2,C)),powerK)+pow((u[i-s]-phihat(z,x,y,q,u,"u","yzqu",j-s,N,eps2,C)),powerK);
		/* vector XYZQ */
		disx_xyzq += pow((x[i-s]-phihat(z,x,y,q,u,"x","xyzqu",j-s,N,eps2,C)),powerK);
		disy_xyzq += pow((y[i-s]-phihat(z,x,y,q,u,"y","xyzqu",j-s,N,eps2,C)),powerK);
		disz_xyzq += pow((z[i-s]-phihat(z,x,y,q,u,"z","xyzqu",j-s,N,eps2,C)),powerK);
		disq_xyzq += pow((q[i-s]-phihat(z,x,y,q,u,"q","xyzqu",j-s,N,eps2,C)),powerK)+pow((u[i-s]-phihat(z,x,y,q,u,"u","xyzqu",j-s,N,eps2,C)),powerK);
      }
      /*if(disy_yq==disy_xyq && disy_xyq==disy_yzq && disy_yzq==disy_xyzq) puts("Tak!");*/
	  /*if(disq_yq==disq_xyq && disq_xyq==disq_yzq && disq_yzq==disq_xyzq) puts("Tak!");*/
	  /*if(disx_xyq==disx_xyzq) puts("Tak!");*/
	  /*if(disz_yzq==disz_xyzq) puts("Tak!");*/


      Cyq[i]+=exp((-Const*(disy_yq+disq_yq))/(pow(epsilon,powerK)));

      Cxyq[i]+=exp((-Const*(disx_xyq+disq_xyq+disy_xyq))/(pow(epsilon,powerK)));

      Cyzq[i]+=exp((-Const*(disz_yzq+disq_yzq+disy_yzq))/(pow(epsilon,powerK)));

      Cxyzq[i]+=exp((-Const*(disx_xyzq+disq_xyzq+disy_xyzq+disz_xyzq))/(pow(epsilon,powerK)));

    }   /* end loop over j */


    Cyq[i] /= (double)(n);
    Cxyq[i] /= (double)(n);
    Cyzq[i] /= (double)(n);
    Cxyzq[i] /= (double)(n);

    h[i] += 2.0*(Cxyzq[i]*Cyq[i] - Cxyq[i]*Cyzq[i])/6.0;

    for (j=mmax;j!=N;j++)
    if (j!=i)
    {
     disx = disy = disq = disz = 0.0;
     disx_xyq = disx_xyzq = disz_yzq = disz_xyzq = disy_yq = disy_xyq = disy_yzq = disy_xyzq = disq_yq = disq_xyq = disq_yzq = disq_xyzq = 0.0;
     for (s=1;s!=m+1;s++) {
		/* vector YQ */
		disy_yq += pow((y[i-s]-phihat(z,x,y,q,u,"y","yqu",j-s,N,eps2,C)),powerK);
		disq_yq += pow((q[i-s]-phihat(z,x,y,q,u,"q","yqu",j-s,N,eps2,C)),powerK)+pow((u[i-s]-phihat(z,x,y,q,u,"u","yqu",j-s,N,eps2,C)),powerK);
		/* vector XYQ */
		disx_xyq += pow((x[i-s]-phihat(z,x,y,q,u,"x","xyqu",j-s,N,eps2,C)),powerK);
		disy_xyq += pow((y[i-s]-phihat(z,x,y,q,u,"y","xyqu",j-s,N,eps2,C)),powerK);
		disq_xyq += pow((q[i-s]-phihat(z,x,y,q,u,"q","xyqu",j-s,N,eps2,C)),powerK)+pow((u[i-s]-phihat(z,x,y,q,u,"u","xyqu",j-s,N,eps2,C)),powerK);
		/* vector YZQ */
		disy_yzq += pow((y[i-s]-phihat(z,x,y,q,u,"y","yzqu",j-s,N,eps2,C)),powerK);
		disz_yzq += pow((z[i-s]-phihat(z,x,y,q,u,"z","yzqu",j-s,N,eps2,C)),powerK);
		disq_yzq += pow((q[i-s]-phihat(z,x,y,q,u,"q","yzqu",j-s,N,eps2,C)),powerK)+pow((u[i-s]-phihat(z,x,y,q,u,"u","yzqu",j-s,N,eps2,C)),powerK);
		/* vector XYZQ */
		disx_xyzq += pow((x[i-s]-phihat(z,x,y,q,u,"x","xyzqu",j-s,N,eps2,C)),powerK);
		disy_xyzq += pow((y[i-s]-phihat(z,x,y,q,u,"y","xyzqu",j-s,N,eps2,C)),powerK);
		disz_xyzq += pow((z[i-s]-phihat(z,x,y,q,u,"z","xyzqu",j-s,N,eps2,C)),powerK);
		disq_xyzq += pow((q[i-s]-phihat(z,x,y,q,u,"q","xyzqu",j-s,N,eps2,C)),powerK)+pow((u[i-s]-phihat(z,x,y,q,u,"u","xyzqu",j-s,N,eps2,C)),powerK);
      }

	IYQij=exp((-Const*(disy_yq+disq_yq))/(pow(epsilon,powerK)));

      IXYQij=exp((-Const*(disx_xyq+disy_xyq+disq_xyq))/(pow(epsilon,powerK)));

      IYZQij=exp((-Const*(disz_yzq+disy_yzq+disq_yzq))/(pow(epsilon,powerK)));

      IXYZQij=exp((-Const*(disx_xyzq+disy_xyzq+disz_xyzq+disq_xyzq))/(pow(epsilon,powerK)));

      h[j] += 2.0*(Cyq[i]*IXYZQij - Cyzq[i]*IXYQij)/(double)(6*n);
      h[j] += 2.0*(Cxyzq[i]*IYQij - Cxyq[i]*IYZQij)/(double)(6*n);

    }   /* end second loop over j */

    T2 += Cxyzq[i]*Cyq[i] - Cyzq[i]*Cxyq[i];

  } /* end loop over i */


/* Alternatively one might determine T2 as the sum of the h[i]

  for (i=mmax;i!=N;i++)
    T2 += h[i];
*/

  T2 = T2*(n-1)/(n*(n-2));

  for (i=mmax;i!=N;i++)
  {
    h[i] -= T2;
  }

  free(Cyq); free(Cxyq); free(Cyzq); free(Cxyzq);
}

void InsertionSort(double *X, int *S, int M)
{
    int i, *I;
    int j;
    int r;
    double R;

    I= (int*) malloc (M*sizeof(int));

    for (i=0;i<M;i++)
      I[i]=i;

    for (i=1; i<M; i++)
      {
        R = X[i];
        r = i;
	for (j=i-1; (j>=0) && (X[j]>R); j--)
        {
	  X[j+1] = X[j];
          I[j+1] = I[j];
        }
	X[j+1] = R;
        I[j+1] = r;
      }
    for (i=0; i<M; i++)
      S[I[i]]=i;

}


void  uniform (double *X, int M)
{
  int *I, i;

  I = (int*) malloc (M*sizeof(int));
  InsertionSort(X, I, M);

  for (i=0;i<M;i++)
    X[i] = (double) I[i]/M*3.464101615;        // to make unit variance

}


/* normalize the time series to unit std. dev. */

void normalise(double *x, int N)
{
  int i;
  double mean=0.0, var=0.0;

  for (i=0;i!=N;i++)
  {
    mean += x[i];
    var += x[i]*x[i];
  }

  mean /= (double)(N);
  var /= (double)(N);
  var -= mean*mean;

  for (i=0;i!=N;i++)
    x[i] = (x[i]-mean)/sqrt(var);

  return;
}

void testcaus(double *z, double *x, double *y, double *q, double *u, int N, double epsilon, int m, int U)
{

  int i, k, mmax;
  double VT2;

  mmax=m;

  if (U<=0)
  {
    normalise(z, N);
    normalise(x, N);
    normalise(y, N);
    normalise(q, N);
    normalise(u, N);
  }
  else
  {
    uniform(z, N);
    uniform(x, N);
    uniform(y, N);
    uniform(q, N);
    uniform(u, N);
  }

  redunGaussian(z,x,y,q,u,N,m,m,epsilon);

  /* determine autocovariance of h[i] */

  for (k=0;k!=K;k++)
  {
    cov[k] = 0.0;

    for (i=mmax+k;i!=N;i++)
      cov[k] += h[i]*h[i-k];

    cov[k] /= (double)(n-k);
  }

  VT2=0.0;

/* variance of T2 */

  for (k=0;k!=K;k++)
    VT2 += 9.0*ohm[k]*cov[k];

  T2_TVAL = T2*sqrt(n)/sqrt(VT2);

  if (T2_TVAL>0)
     p_T2 = 0.5 - 0.5*erf(T2_TVAL/sqrt(2.0));
  else
     p_T2 = 0.5 + 0.5*erf(-T2_TVAL/sqrt(2.0));

}


int main()
{
  char filename1[128], filename2[128], filename3[128], filename4[128], filename[128];
  double z[Ndat], x[Ndat], y[Ndat], q[Ndat], u[Ndat], tmp, epsilon=EPS;
  int i, j, k, m, N;
  //  long seed;

  for (i=0;i<N;i++)
    x[i]=0;
   for (i=0;i<N;i++)
    y[i]=0;
   for (i=0;i<N;i++)
    q[i]=0;
   for (i=0;i<N;i++)
    z[i]=0;
   for (i=0;i<N;i++)
    u[i]=0;

  /* reading files */

  FILE *infil, *outfil;

  printf("Input file (X): "); scanf("%s", filename1);

  if ( (infil=fopen(filename1,"r")) == NULL)
  {
    fprintf(stderr,"Error: unable to open file. Exiting...\n");
    exit(1);
  }

  i = 0;
  while (fscanf(infil,"%lf", &tmp) != EOF)
  {
      x[i] = tmp;
      i++;
  }

  N = i;

  printf("%d data read\n", i);

  fclose(infil);

  printf("Input file (Y): "); scanf("%s", filename2);

  if ( (infil=fopen(filename2,"r")) == NULL)
  {
    fprintf(stderr,"Error: unable to open file. Exiting...\n");
    exit(1);
  }

  i = 0;

  while (fscanf(infil,"%lf", &tmp) != EOF)
  {
      y[i] = tmp;
      i++;
  }

  printf("%d data read\n", i);

  fclose(infil);

  if (i!=N)
  {
    fprintf(stderr,"Error: data number mismatch (X and Y). Exiting...\n");
    exit(1);
  }

  printf("Input file (Q): "); scanf("%s", filename3);

  if ( (infil=fopen(filename3,"r")) == NULL)
  {
    fprintf(stderr,"Error: unable to open file. Exiting...\n");
    exit(1);
  }

  i = 0;

  while (fscanf(infil,"%lf", &tmp) != EOF)
  {
      q[i] = tmp;
      i++;
  }

  printf("%d data read\n", i);

  printf("Input file (U): "); scanf("%s", filename4);

  if ( (infil=fopen(filename4,"r")) == NULL)
  {
    fprintf(stderr,"Error: unable to open file. Exiting...\n");
    exit(1);
  }

  i = 0;

  while (fscanf(infil,"%lf", &tmp) != EOF)
  {
      u[i] = tmp;
      i++;
  }

  printf("%d data read\n", i);

  fclose(infil);

  if (i!=N)
  {
    fprintf(stderr,"Error: data number mismatch (Q and X/Y). Exiting...\n");
    exit(1);
  }

  A = (double **) malloc(4*sizeof(double *));
  for (i=0;i!=4;i++)
    A[i] = (double *) malloc(N*sizeof(double));

  h = (double *) malloc(N*sizeof(double));

  K = (int)(sqrt(sqrt(N)));
  ohm = (double *) malloc(K*sizeof(double));
  cov = (double *) malloc(K*sizeof(double));

  ohm[0] = 1.0;
  for (k=1;k<K;k++)
    ohm[k] = 2.0*(1.0-k/(double)(K));

  /* Generate output file name */

  //Generate output file name
  i=0;

  while (filename1[i]!='\0' && filename1[i]!='.')
  {
    filename[i]=filename1[i];
    i++;
  }
  filename1[i]='\0';
  filename[i]='_';

  j=0;

  while (filename2[j]!='\0' && filename2[j]!='.')
  {
    filename[i+j+1]=filename2[j];
    j++;
  }

  filename2[j]='\0';
  filename[i+j+1]='.';  filename[i+j+2]='o'; filename[i+j+3]='u';
  filename[i+j+4]='t'; filename[i+j+5]='\0';


  if ( (outfil=fopen(filename,"w")) == NULL)
  {
    fprintf(stderr,"Error: unable to open file for writing.\
	Exiting...\n");
    exit(1);
  }

  fprintf(outfil,"file 1: %s\n", filename1);
  fprintf(outfil,"file 2: %s\n", filename2);
  fprintf(outfil,"series length: %d\n",N);
  fprintf(outfil,"epsilon=%f\n",epsilon);

  fprintf(stderr,"epsilon=%f\n",epsilon);
  fprintf(outfil,"epsilon=%f\n",epsilon);


  fprintf(stderr,"\n%s does not granger cause %s (corrected for the presence of %s and %s)\n", filename1, filename2, filename3, filename4);
  fprintf(outfil,"\n%s does not granger cause %s (corrected for the presence of %s and %s)\n", filename1, filename2, filename3, filename4);

  /*generate vector z*/
  for (i=1;i!=N;i++) {
	z[i-1]=y[i];
  }

  for (m=1;m<=Mmax;m++)
  {
    testcaus(z,x,y,q,u,N,epsilon,1,0); /*data sharpening normal distribution adjustment*/

    fprintf(stderr,"lX=lY=lQ=lU=%d, p_T2=%1.6f T_T2=%1.6f\n",m,p_T2,T2_TVAL);
    fprintf(outfil,"lX=lY=lQ=lU=%d, p_T2=%1.6f T_T2=%1.6f\n",m,p_T2,T2_TVAL);
  }

  fprintf(stderr,"\n%s does not granger cause %s (corrected for the presence of %s and %s) - UNIF \n", filename1, filename2, filename3, filename4);
  fprintf(outfil,"\n%s does not granger cause %s (corrected for the presence of %s and %s) - UNIF \n", filename1, filename2, filename3, filename4);

  for (m=1;m<=Mmax;m++)
  {
    testcaus(z,x,y,q,u,N,epsilon,1,1); /*data sharpening normal distribution adjustment*/

    fprintf(stderr,"lX=lY=lQ=lU=%d, p_T2=%1.6f T_T2=%1.6f\n",m,p_T2,T2_TVAL);
    fprintf(outfil,"lX=lY=lQ=lU=%d, p_T2=%1.6f T_T2=%1.6f\n",m,p_T2,T2_TVAL);
  }


  fprintf(stderr,"\n%s does not granger cause %s (corrected for the presence of %s and %s)\n", filename2, filename1, filename3, filename4);
  fprintf(outfil,"\n%s does not granger cause %s (corrected for the presence of %s and %s)\n", filename2, filename1, filename3, filename4);


  /*generate vector z*/
  for (i=1;i!=N;i++) {
	z[i-1]=x[i];
  }
  for (m=1;m<=Mmax;m++)
  {
    testcaus(z,y,x,q,u,N,epsilon,1,0); /*data sharpening normal distribution adjustment*/

    fprintf(stderr,"lX=lY=lQ=lU=%d, p_T2=%1.6f T_T2=%1.6f\n",m,p_T2,T2_TVAL);
    fprintf(outfil,"lX=lY=lQ=lU=%d, p_T2=%1.6f T_T2=%1.6f\n",m,p_T2,T2_TVAL);
  }

  fprintf(stderr,"\n%s does not granger cause %s (corrected for the presence of %s and %s) - UNIF \n", filename2, filename1, filename3, filename4);
  fprintf(outfil,"\n%s does not granger cause %s (corrected for the presence of %s and %s) - UNIF \n", filename2, filename1, filename3, filename4);


  for (m=1;m<=Mmax;m++)
  {
    testcaus(z,y,x,q,u,N,epsilon,1,1); /*data sharpening normal distribution adjustment*/

    fprintf(stderr,"lX=lY=lQ=lU=%d, p_T2=%1.6f T_T2=%1.6f\n",m,p_T2,T2_TVAL);
    fprintf(outfil,"lX=lY=lQ=lU=%d, p_T2=%1.6f T_T2=%1.6f\n",m,p_T2,T2_TVAL);
  }

  fclose(outfil);

  return(0);
}
